/** ZCT/searchu.c--Created 102906 **/

/** Copyright 2008 Zach Wegner **/
/*
 * This file is part of ZCT.
 *
 * ZCT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ZCT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ZCT.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "zct.h"
#include "functions.h"
#include "globals.h"
#include "eval.h"
#include "smp.h"

/**
heuristic():
Updates history, killer, and counter tables based on a move that failed high.
Created 092906; last modified 081007
**/
void heuristic(MOVE best_move, SEARCH_BLOCK *sb)
{
	int x;

	history_table[board.side_tm][MOVE_KEY(best_move)] += sb->depth * sb->depth;
	history_counter = MAX(history_table[board.side_tm][MOVE_KEY(best_move)], history_counter);

	if (history_counter > 1 << 16)
	{
		for (x = 0; x < 2 * 4096; x++) /* This is not recommended ;) */
			history_table[0][x] >>= 6;
		history_counter >>= 6;
	}
	if (board.color[MOVE_TO(best_move)] == EMPTY)
	{
		killer_move[sb->ply][1] = MOVE_COMPARE(best_move);
		counter_move[board.side_tm][MOVE_KEY((sb - 1)->move)][1] = MOVE_COMPARE(best_move);
	}
}

/**
heuristic_pv():
Updates killer tables when the pv changes at a pv node.
Created 110606; last modified 081007
**/
void heuristic_pv(MOVE best_move, SEARCH_BLOCK *sb)
{
	if (board.color[MOVE_TO(best_move)] == EMPTY)
	{
		killer_move[sb->ply][0] = MOVE_COMPARE(best_move);
		counter_move[board.side_tm][MOVE_KEY((sb - 1)->move)][0] = MOVE_COMPARE(best_move);
	}
}

/**
moves_are_connected():
This function is used primarily in LMR, for testing if a reduced move actually contained a threat. The
function tests if the move that refuted null move was connected to the first move. This idea is from Tord Romstad.
Created 101707; last modified 101807
**/
BOOL moves_are_connected(MOVE move_1, MOVE move_2)
{
//return FALSE;
	if (move_1 == NULL_MOVE || move_2 == NULL_MOVE)
		return FALSE;
	if (MOVE_TO(move_1) == MOVE_FROM(move_2))
		return TRUE;
	if (MOVE_FROM(move_1) == MOVE_TO(move_2))
		return TRUE;
	if (MASK(MOVE_FROM(move_1)) & in_between_bb[MOVE_FROM(move_2)][MOVE_TO(move_2)])
		return TRUE;
	return FALSE;
}

/**
move_is_futile():
This function tests whether a move should be pruned using Ernst Heinz's Futility Pruning. The idea
is that, at horizon nodes, if a move cannot bring the material score (plus a margin) to alpha, there
is no point in even making it, because the next ply in quiescence will have a stand-pat fail high.
Created 101707; last modified 101707
**/
BOOL move_is_futile(SEARCH_BLOCK *sb, MOVE move)
{
//	return FALSE;
	return (sb->moves > 0 && sb->depth + sb->extension <= PLY &&
		material_balance() + piece_value[board.piece[MOVE_TO(sb->move)]] + 100 < sb->alpha);
}

/**
should_reduce():
This function test whether a move should be reduced by Late Move Reductions. There aren't many conditions
at the moment.
Created 101707; last modified 021008
**/
BOOL should_reduce(SEARCH_BLOCK *sb, MOVE move)
{
	return (sb->moves > 2 && sb->extension == 0 && sb->depth >= 2 * PLY &&
		/* A move must be below a certain percentage of the max history score. */
1);//		MOVE_SCORE(sb->move) < 2000 + (history_threshold * 64 / 100));
}

/**
null_reduction():
Determine the amount of reduction to apply to null move searches.
Created 103006; last modified 110606
**/
int null_reduction(SEARCH_BLOCK *sb)
{
	return 3 * PLY;
}

/**
extend():
Calculates the search extension to be awarded to the given move. This is based on various information
such as whether the move checks, if it is a threat or not, etc.
Created 072007; last modified 072007
**/
int extend(SEARCH_BLOCK *sb)
{
	int extension;

	extension = 0;
	/* check extension */
	if (in_check())
	{
		extension += check_extension;
		check_extensions_done++;
	}
	/* one-reply extension */
	if (sb->check && sb->last_move - sb->first_move == 1)
	{
		extension += one_rep_extension;
		one_rep_extensions_done++;
	}
	/* threat move extension */
	if (sb->threat)
	{
		extension += threat_extension;
		threat_extensions_done++;
	}
	/* pawn to 7th extension */
	if (board.piece[MOVE_TO(sb->move)] == PAWN && RANK_OF(SQ_FLIP_COLOR(MOVE_TO(sb->move), board.side_ntm)) == RANK_7)
	{
		extension += passed_pawn_extension;
		passed_pawn_extensions_done++;
	}
	extension = MIN(PLY, extension);
	return extension;
}

/**
search_call():
Sets up the internal stack for a search call.
Created 081205; last modified 081007
**/
void search_call(SEARCH_BLOCK *sb, BOOL is_quiesce, int depth, int ply, VALUE alpha, VALUE beta,
	MOVE *first_move, NODE_TYPE node_type, SEARCH_STATE search_state)
{
	if (depth <= 0 && is_quiesce == FALSE)
	{
		depth = 2;
		is_quiesce = TRUE;
	}
	sb->search_state = search_state;
	sb++;
	if (is_quiesce)
		sb->search_state = QSEARCH_START;
	else
		sb->search_state = SEARCH_START;
	sb->depth = depth;
	sb->ply = ply;
	sb->alpha = alpha;
	sb->beta = beta;
	sb->first_move = first_move;
	sb->next_move = first_move;
	sb->last_move = first_move;
	sb->move_made = FALSE;
	sb->select_state = SELECT_HASH_MOVE;
	sb->node_type = node_type;

	sb->hash_move = 0;
	sb->move = NULL_MOVE;
	sb->moves = 0;
	sb->best_score = -MATE;
	board.pv_stack[sb->ply][0] = 0;
	sb->pv_found = 0;
	sb->threat = FALSE;

	sb->id = sb_id++;
}

/**
copy_pv():
Copies the pv from from to to. :)
Created 082906; last modified 091906
**/
void copy_pv(MOVE *to, MOVE *from)
{
	while (*from)
		*to++ = *from++;
	*to = 0;
}

/**
stop_search():
This checks for time and input to see if need to stop searching.
Created 091006; last modified 013008
**/
BOOL stop_search(void)
{
	CMD_RESULT cmd_result;
	BOOL stop = FALSE;
	GAME_ENTRY *current;

	/* Check for input, because we might enter this function more than once before fully stopping. */
	if (input_buffer[0])
		stop = TRUE;
#ifdef SMP
	else if (smp_data->stop)
		stop = TRUE;
#endif
	else if (time_is_up())
		stop = TRUE;
	else if (input_available())
	{
		/* Return to the root position. */
		current = board.game_entry;
		while (board.game_entry > root_entry)
			unmake_move();
		if (engine_state == PONDERING)
			unmake_move();
		read_line(input_buffer, sizeof(input_buffer));
		cmd_result = command(input_buffer);
		/* Command() returns a 2 when we need to exit the search to handle a command. */
		if (cmd_result == CMD_STOP_SEARCH)
			stop = TRUE;
		else if (engine_state != NORMAL && cmd_result == CMD_BAD && input_move(input_buffer, INPUT_CHECK_MOVE))
		{
			if (engine_state == PONDERING && input_move(input_buffer, INPUT_GET_MOVE) == ponder_move)
			{
				make_move(ponder_move);
				engine_state = NORMAL;
				update_clocks();
			}
			else
				stop = TRUE;
		}
		else if (cmd_result == CMD_BAD && protocol != UCI)
			print("Error.\n");
		if (ping && protocol == UCI)
		{
			print("readyok\n");
			ping = 0;
		}
		/* Go back to the position we were in before checking. This is kind of hacky in that it relies on the
			game history after undoing all the moves to be the same. */
		while (board.game_entry < current)
			make_move(board.game_entry->move);
		if (!stop)
			input_buffer[0] = '\0';
	}
	return stop;
}
